from flask import Flask, Blueprint, jsonify

from pythondi import Provider, configure, inject
from .repo import Repo, SQLRepo

bp = Blueprint('home', __name__)


@bp.route('/')
def home():
    usecase = Usecase()
    usecase.repo.get()
    return jsonify({'hello': 'world'})


class Usecase:
    @inject()
    def __init__(self, repo: Repo):
        self.repo = repo


def create_app():
    provider = Provider()
    provider.bind(Repo, SQLRepo)
    configure(provider=provider)
    app = Flask(__name__)
    app.register_blueprint(bp)
    return app


if __name__ == '__main__':
    app = create_app()
    app.run(debug=True)



from pytest import raises

from pythondi import (
    configure,
    configure_after_clear,
    Provider,
    InjectException
)


def test_configure():
    provider = Provider()
    provider.bind(int, str)
    configure(provider=provider)
    from pythondi import _PROVIDER
    assert isinstance(_PROVIDER, Provider)
    assert _PROVIDER.bindings[int] == str


def test_configure_after_clear():
    _PROVIDER = None
    provider = Provider()
    provider.bind(int, str)
    with raises(InjectException):
        configure(provider=provider)

    provider = Provider()
    provider.bind(str, int)
    from pythondi import _PROVIDER
    configure_after_clear(provider=provider)
    assert isinstance(_PROVIDER, Provider)
    assert _PROVIDER.bindings[int] == str




import inspect
import threading
from functools import wraps
from typing import Optional, NoReturn

_PROVIDER = None
_LOCK = threading.RLock()


class InjectException(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class Provider:
    def __init__(self):
        self._bindings = {}

    def bind(self, cls, new_cls) -> None:
        """Binding class to another class"""
        self._bindings[cls] = new_cls

    def unbind(self, cls) -> None:
        """Unbinding class"""
        try:
            self._bindings.pop(cls)
        except KeyError:
            raise InjectException(msg='Unbind exception')

    def clear_bindings(self) -> None:
        """Clear bindings"""
        self._bindings = {}

    @property
    def bindings(self) -> dict:
        """Return current binding classes"""
        return self._bindings


def configure(provider: Provider) -> Optional[NoReturn]:
    """Configure provider"""
    global _PROVIDER

    with _LOCK:
        if _PROVIDER:
            raise InjectException(msg='Already injected')

        _PROVIDER = provider


def configure_after_clear(provider: Provider) -> None:
    """Clear existing provider and configure new provider"""
    global _PROVIDER

    with _LOCK:
        if _PROVIDER:
            clear()

        _PROVIDER = provider


def clear() -> None:
    global _PROVIDER

    with _LOCK:
        _PROVIDER = None


def inject(**params):
    def inner_func(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Case of auto injection
            if params == {}:
                annotations = inspect.getfullargspec(func).annotations
                for k, v in annotations.items():
                    if v in _PROVIDER.bindings:
                        kwargs[k] = _PROVIDER.bindings[v]()
            # Case of manual injection
            else:
                for k, v in params.items():
                    kwargs[k] = v()
            func(*args, **kwargs)
        return wrapper
    return inner_func



from pythondi import inject, Provider, configure_after_clear


class Repo:
    def __init__(self):
        pass


class SQLRepo:
    def __init__(self):
        pass


def test_inject_without_parameter():
    provider = Provider()
    provider.bind(Repo, SQLRepo)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo):
        assert isinstance(repo, SQLRepo)

    func()


def test_inject_with_parameter():
    provider = Provider()
    provider.bind(Repo, SQLRepo)
    configure_after_clear(provider)

    @inject(repo=SQLRepo)
    def func(repo):
        assert isinstance(repo, SQLRepo)

    func()



import abc


class Repo:
    """Interface class"""
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get(self):
        pass


class SQLRepo(Repo):
    """Impl class"""
    def __init__(self):
        pass

    def get(self):
        print('SQLRepo')



from pytest import raises

from pythondi import Provider, InjectException


def test_bind():
    provider = Provider()
    provider.bind(int, str)
    assert provider.bindings[int] == str


def test_unbind():
    provider = Provider()
    with raises(InjectException):
        provider.unbind(cls=int)

    provider.bind(int, str)
    provider.unbind(cls=int)
    assert provider.bindings == {}


def test_clear_bindings():
    provider = Provider()
    provider.bind(int, str)
    provider.bind(list, dict)
    provider.clear_bindings()
    assert provider.bindings == {}



from pythondi import Provider, configure, configure_after_clear, inject
from .repo import Repo, SQLRepo


class Usecase:
    @inject()
    def __init__(self, repo: Repo):
        self.repo = repo


if __name__ == '__main__':
    # Init provider
    provider = Provider()

    # Bind `Impl` class to `Interface` class
    provider.bind(Repo, SQLRepo)

    # Inject with configure
    configure(provider=provider)

    # Or if you want to fresh inject, use `configure_after_clear`
    configure_after_clear(provider=provider)

    # Init class without arguments
    u = Usecase()
